@@ -1,3 +1,24 @@
+// Copyright Joyent, Inc. and other Node contributors.
+//
+// Permission is hereby granted, free of charge, to any person obtaining a
+// copy of this software and associated documentation files (the
+// "Software"), to deal in the Software without restriction, including
+// without limitation the rights to use, copy, modify, merge, publish,
+// distribute, sublicense, and/or sell copies of the Software, and to permit
+// persons to whom the Software is furnished to do so, subject to the
+// following conditions:
+//
+// The above copyright notice and this permission notice shall be included
+// in all copies or substantial portions of the Software.
+//
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
+// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
+// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
+// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
+// USE OR OTHER DEALINGS IN THE SOFTWARE.
+
 var util = require('util');
 
 var binding = process.binding('fs');
@@ -219,7 +240,12 @@ fs.read = function(fd, buffer, offset, length, position, callback) {
     };
   }
 
-  binding.read(fd, buffer, offset, length, position, callback || noop);
+  function wrapper(err, bytesRead) {
+    // Retain a reference to buffer so that it can't be GC'ed too soon.
+    callback && callback(err, bytesRead || 0, buffer);
+  }
+
+  binding.read(fd, buffer, offset, length, position, wrapper);
 };
 
 fs.readSync = function(fd, buffer, offset, length, position) {
@@ -264,7 +290,12 @@ fs.write = function(fd, buffer, offset, length, position, callback) {
     return;
   }
 
-  binding.write(fd, buffer, offset, length, position, callback || noop);
+  function wrapper(err, written) {
+    // Retain a reference to buffer so that it can't be GC'ed too soon.
+    callback && callback(err, written || 0, buffer);
+  }
+
+  binding.write(fd, buffer, offset, length, position, wrapper);
 };
 
 fs.writeSync = function(fd, buffer, offset, length, position) {
@@ -807,7 +838,9 @@ ReadStream.prototype.setEncoding = function(encoding) {
 
 ReadStream.prototype._read = function() {
   var self = this;
-  if (!self.readable || self.paused) return;
+  if (!self.readable || self.paused || self.reading) return;
+
+  self.reading = true;
 
   if (!pool || pool.length - pool.used < kMinPoolSpace) {
     // discard the old pool. Can't add to the free list because
@@ -833,6 +866,7 @@ ReadStream.prototype._read = function() {
   }
 
   function afterRead(err, bytesRead) {
+    self.reading = false;
     if (err) {
       self.emit('error', err);
       self.readable = false;
@@ -1051,7 +1085,15 @@ WriteStream.prototype.write = function(data) {
   return false;
 };
 
-WriteStream.prototype.end = function(cb) {
+WriteStream.prototype.end = function(data, encoding, cb) {
+  if (typeof(data) === 'function') {
+    cb = data;
+  } else if (typeof(encoding) === 'function') {
+    cb = encoding;
+    this.write(data);
+  } else if (arguments.length > 0) {
+    this.write(data, encoding);
+  }
   this.writable = false;
   this._queue.push([fs.close, cb]);
   this.flush();
