@@ -1,3 +1,24 @@
+// Copyright Joyent, Inc. and other Node contributors.
+//
+// Permission is hereby granted, free of charge, to any person obtaining a
+// copy of this software and associated documentation files (the
+// "Software"), to deal in the Software without restriction, including
+// without limitation the rights to use, copy, modify, merge, publish,
+// distribute, sublicense, and/or sell copies of the Software, and to permit
+// persons to whom the Software is furnished to do so, subject to the
+// following conditions:
+//
+// The above copyright notice and this permission notice shall be included
+// in all copies or substantial portions of the Software.
+//
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
+// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
+// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
+// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
+// USE OR OTHER DEALINGS IN THE SOFTWARE.
+
 var util = require('util');
 var net = require('net');
 var stream = require('stream');
@@ -261,6 +282,7 @@ IncomingMessage.prototype._addHeaderLine = function(field, value) {
     case 'accept-language':
     case 'connection':
     case 'cookie':
+    case 'pragma':
       if (field in dest) {
         dest[field] += ', ' + value;
       } else {
@@ -730,6 +752,7 @@ OutgoingMessage.prototype._flush = function() {
     // This is a queue to the server or client to bring in the next this.
     this._finish();
   } else if (ret) {
+    // This is necessary to prevent https from breaking
     this.emit('drain');
   }
 };
@@ -839,9 +862,11 @@ ServerResponse.prototype.writeHeader = function() {
 };
 
 
-function ClientRequest(options) {
+function ClientRequest(options, defaultPort) {
   OutgoingMessage.call(this);
 
+  if (!defaultPort) defaultPort = 80;
+
   var method = this.method = (options.method || 'GET').toUpperCase();
   this.path = options.path || '/';
 
@@ -856,7 +881,11 @@ function ClientRequest(options) {
     }
     // Host header set by default.
     if (options.host && !this.getHeader('host')) {
-      this.setHeader("Host", options.host);
+      var hostHeader = options.host;
+      if (options.port && +options.port !== defaultPort) {
+        hostHeader += ':' + options.port;
+      }
+      this.setHeader("Host", hostHeader);
     }
   }
 
@@ -951,6 +980,7 @@ function connectionListener(socket) {
     while (incoming.length) {
       var req = incoming.shift();
       req.emit('aborted');
+      req.emit('close');
     }
     // abort socket._httpMessage ?
   }
@@ -1086,7 +1116,7 @@ exports._connectionListener = connectionListener;
 function Agent(options) {
   this.options = options;
   this.host = options.host;
-  this.port = options.port || 80;
+  this.port = options.port || this.defaultPort;
 
   this.queue = [];
   this.sockets = [];
@@ -1098,11 +1128,11 @@ exports.Agent = Agent;
 
 Agent.defaultMaxSockets = 5;
 
-
+Agent.prototype.defaultPort = 80;
 Agent.prototype.appendMessage = function(options) {
   var self = this;
 
-  var req = new ClientRequest(options);
+  var req = new ClientRequest(options, this.defaultPort);
   this.queue.push(req);
   req._queue = this.queue;
 
@@ -1175,7 +1205,6 @@ Agent.prototype._establishNewConnection = function() {
     parser.finish();
     socket.destroy();
     self._removeSocket(socket);
-    parsers.free(parser);
     self._cycle();
   });
 
@@ -1233,10 +1262,15 @@ Agent.prototype._establishNewConnection = function() {
     // has a request but no response and it never emitted an error event:
     // THEN we need to trigger it manually.
     // There must be a better way to do this.
-    if (socket._httpMessage &&
-        !socket._httpMessage.res &&
-        !socket._httpMessage._hadError) {
-      socket._httpMessage.emit('error', new Error('socket hang up'));
+    if (socket._httpMessage) {
+      if (socket._httpMessage.res) {
+        socket._httpMessage.res.emit('aborted');
+        socket._httpMessage.res.emit('close');
+      } else {
+        if (!socket._httpMessage._hadError) {
+          socket._httpMessage.emit('error', new Error('socket hang up'));
+        }
+      }
     }
 
     self._removeSocket(socket);
@@ -1281,8 +1315,10 @@ Agent.prototype._establishNewConnection = function() {
       // All that should be required for keep-alive is to not reconnect,
       // but outgoingFlush instead.
       if (!req.shouldKeepAlive) {
-        debug('AGENT socket.end()');
-        if (socket.writable) socket.end();
+        if (socket.writable) {
+          debug('AGENT socket.destroySoon()');
+          socket.destroySoon();
+        }
         assert(!socket.writable);
       } else {
         debug('AGENT socket keep-alive');
@@ -1340,6 +1376,7 @@ Agent.prototype._cycle = function() {
       first._queue = null;
 
       first.assignSocket(socket);
+      httpSocketSetup(socket);
       self._cycle(); // try to dispatch another
       return;
     }
