@@ -1,3 +1,24 @@
+// Copyright Joyent, Inc. and other Node contributors.
+//
+// Permission is hereby granted, free of charge, to any person obtaining a
+// copy of this software and associated documentation files (the
+// "Software"), to deal in the Software without restriction, including
+// without limitation the rights to use, copy, modify, merge, publish,
+// distribute, sublicense, and/or sell copies of the Software, and to permit
+// persons to whom the Software is furnished to do so, subject to the
+// following conditions:
+//
+// The above copyright notice and this permission notice shall be included
+// in all copies or substantial portions of the Software.
+//
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
+// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
+// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
+// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
+// USE OR OTHER DEALINGS IN THE SOFTWARE.
+
 var NativeModule = require('native_module');
 var Script = process.binding('evals').Script;
 var runInThisContext = Script.runInThisContext;
@@ -202,7 +223,10 @@ Module._resolveLookupPaths = function(request, parent) {
   var start = request.substring(0, 2);
   if (start !== './' && start !== '..') {
     var paths = Module._paths;
-    if (parent) paths = paths.concat(parent.paths);
+    if (parent) {
+      if (!parent.paths) parent.paths = [];
+      paths = parent.paths.concat(paths);
+    }
     return [request, paths];
   }
 
@@ -211,7 +235,7 @@ Module._resolveLookupPaths = function(request, parent) {
     // make require('./path/to/foo') work - normally the path is taken
     // from realpath(__filename) but with eval there is no filename
     var mainPaths = ['.'].concat(Module._paths);
-    mainPaths = mainPaths.concat(Module._nodeModulePaths('.'));
+    mainPaths = Module._nodeModulePaths('.').concat(mainPaths);
     return [request, mainPaths];
   }
 
@@ -270,7 +294,13 @@ Module._load = function(request, parent, isMain) {
   }
 
   Module._cache[filename] = module;
-  module.load(filename);
+  try {
+    module.load(filename);
+  } catch (err) {
+    delete Module._cache[filename];
+    throw err;
+  }
+
   return module.exports;
 };
 
