@@ -1,3 +1,24 @@
+// Copyright Joyent, Inc. and other Node contributors.
+//
+// Permission is hereby granted, free of charge, to any person obtaining a
+// copy of this software and associated documentation files (the
+// "Software"), to deal in the Software without restriction, including
+// without limitation the rights to use, copy, modify, merge, publish,
+// distribute, sublicense, and/or sell copies of the Software, and to permit
+// persons to whom the Software is furnished to do so, subject to the
+// following conditions:
+//
+// The above copyright notice and this permission notice shall be included
+// in all copies or substantial portions of the Software.
+//
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
+// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
+// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
+// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
+// USE OR OTHER DEALINGS IN THE SOFTWARE.
+
 var util = require('util');
 var events = require('events');
 var stream = require('stream');
@@ -118,7 +139,7 @@ function setImplmentationMethods(self) {
       // emitting the same value that we see now. Otherwise, we can end up
       // calling emit() after recvMsg() has been called again and end up
       // emitting null (or another FD).
-      if (recvMsg.fd !== null) {
+      if (typeof recvMsg.fd === 'number') {
         var fd = recvMsg.fd;
         process.nextTick(function() {
           self.emit('fd', fd);
@@ -172,7 +193,7 @@ function initSocket(self) {
   self._readWatcher = ioWatchers.alloc();
   self._readWatcher.socket = self;
   self._readWatcher.callback = onReadable;
-  self.readable = false;
+  self.readable = self.destroyed = false;
 
   // Queue of buffers and string that need to be written to socket.
   self._writeQueue = [];
@@ -251,16 +272,16 @@ Object.defineProperty(Socket.prototype, 'readyState', {
     if (this._connecting) {
       return 'opening';
     } else if (this.readable && this.writable) {
-      assert(typeof this.fd == 'number');
+      assert(typeof this.fd === 'number');
       return 'open';
     } else if (this.readable && !this.writable) {
-      assert(typeof this.fd == 'number');
+      assert(typeof this.fd === 'number');
       return 'readOnly';
     } else if (!this.readable && this.writable) {
-      assert(typeof this.fd == 'number');
+      assert(typeof this.fd === 'number');
       return 'writeOnly';
     } else {
-      assert(typeof this.fd != 'number');
+      assert(typeof this.fd !== 'number');
       return 'closed';
     }
   }
@@ -526,6 +547,8 @@ Socket.prototype.setEncoding = function(encoding) {
 
 
 function doConnect(socket, port, host) {
+  if (socket.destroyed) return;
+
   timers.active(socket);
 
   try {
@@ -665,7 +688,7 @@ Socket.prototype._onReadable = function() {
 Socket.prototype.connect = function() {
   var self = this;
   initSocket(self);
-  if (self.fd) throw new Error('Socket already opened');
+  if (typeof self.fd === 'number') throw new Error('Socket already opened');
   if (!self._readWatcher) throw new Error('No readWatcher');
 
   timers.active(this);
@@ -723,7 +746,7 @@ Socket.prototype.setKeepAlive = function(enable, time) {
 Socket.prototype.setTimeout = function(msecs, callback) {
   if (msecs > 0) {
     timers.enroll(this, msecs);
-    if (this.fd) { timers.active(this); }
+    if (typeof this.fd === 'number') { timers.active(this); }
     if (callback) {
       this.once('timeout', callback);
     }
@@ -739,7 +762,9 @@ Socket.prototype.pause = function() {
 
 
 Socket.prototype.resume = function() {
-  if (this.fd === null) throw new Error('Cannot resume() closed Socket.');
+  if (typeof this.fd !== 'number') {
+    throw new Error('Cannot resume() closed Socket.');
+  }
   if (this._readWatcher) {
     this._readWatcher.stop();
     this._readWatcher.set(this.fd, true, false);
@@ -788,12 +813,12 @@ Socket.prototype.destroy = function(exception) {
 
   timers.unenroll(this);
 
-  if (this.server) {
+  if (this.server && !this.destroyed) {
     this.server.connections--;
   }
 
   // FIXME Bug when this.fd == 0
-  if (typeof this.fd == 'number') {
+  if (typeof this.fd === 'number') {
     debug('close ' + this.fd);
     close(this.fd);
     this.fd = null;
@@ -802,6 +827,8 @@ Socket.prototype.destroy = function(exception) {
       self.emit('close', exception ? true : false);
     });
   }
+
+  this.destroyed = true;
 };
 
 
@@ -875,7 +902,7 @@ function Server(/* [ options, ] listener */) {
       self.watcher.stop();
     }
 
-    while (self.fd) {
+    while (typeof self.fd === 'number') {
       try {
         var peerInfo = accept(self.fd);
       } catch (e) {
@@ -984,7 +1011,7 @@ Server.prototype._rejectPending = function() {
 // server.listen(8000, '192.168.1.2');
 Server.prototype.listen = function() {
   var self = this;
-  if (self.fd) throw new Error('Server already opened');
+  if (typeof self.fd === 'number') throw new Error('Server already opened');
 
   var lastArg = arguments[arguments.length - 1];
   if (typeof lastArg == 'function') {
@@ -1039,7 +1066,7 @@ Server.prototype.listen = function() {
 };
 
 Server.prototype.listenFD = function(fd, type) {
-  if (this.fd) {
+  if (typeof this.fd === 'number') {
     throw new Error('Server already opened');
   }
 
@@ -1074,7 +1101,7 @@ Server.prototype._doListen = function() {
     // It could be that server.close() was called between the time the
     // original listen command was issued and this. Bail if that's the case.
     // See test/simple/test-net-eaddrinuse.js
-    if (typeof self.fd != 'number') return;
+    if (typeof self.fd !== 'number') return;
 
     try {
       listen(self.fd, self._backlog || 128);
@@ -1096,7 +1123,7 @@ Server.prototype.address = function() {
 
 Server.prototype.close = function() {
   var self = this;
-  if (!self.fd) throw new Error('Not running');
+  if (typeof self.fd !== 'number') throw new Error('Not running');
 
   self.watcher.stop();
 
