@@ -1,3 +1,24 @@
+// Copyright Joyent, Inc. and other Node contributors.
+//
+// Permission is hereby granted, free of charge, to any person obtaining a
+// copy of this software and associated documentation files (the
+// "Software"), to deal in the Software without restriction, including
+// without limitation the rights to use, copy, modify, merge, publish,
+// distribute, sublicense, and/or sell copies of the Software, and to permit
+// persons to whom the Software is furnished to do so, subject to the
+// following conditions:
+//
+// The above copyright notice and this permission notice shall be included
+// in all copies or substantial portions of the Software.
+//
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
+// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
+// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
+// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
+// USE OR OTHER DEALINGS IN THE SOFTWARE.
+
 var crypto = require('crypto');
 var util = require('util');
 var net = require('net');
@@ -147,30 +168,32 @@ CryptoStream.prototype.getCipher = function(err) {
 
 
 CryptoStream.prototype.end = function(d) {
-  if (!this.writable) {
-    throw new Error('CryptoStream is not writable');
+  if (this.writable) {
+    if (this.pair._done) return;
+
+    if (d) {
+      this.write(d);
+    }
+
+    this._pending.push(END_OF_FILE);
+    this._pendingCallbacks.push(null);
+
+    // If this is an encrypted stream then we need to disable further 'data'
+    // events.
+
+    this.writable = false;
+
+    this.pair._cycle();
   }
-
-  if (this.pair._done) return;
-
-  if (d) {
-    this.write(d);
-  }
-
-  this._pending.push(END_OF_FILE);
-  this._pendingCallbacks.push(null);
-
-  // If this is an encrypted stream then we need to disable further 'data'
-  // events.
-
-  this.writable = false;
-
-  this.pair._cycle();
 };
 
 
 CryptoStream.prototype.destroySoon = function(err) {
-  return this.end();
+  if (this.writable) {
+    this.end();
+  } else {
+    this.destroy();
+  }
 };
 
 
@@ -218,7 +241,7 @@ CryptoStream.prototype._push = function() {
   while (this._writeState == true) {
     var bytesRead = 0;
     var chunkBytes = 0;
-    var pool = new Buffer(4096); // alloc every time?
+    var pool = new Buffer(16 * 4096); // alloc every time?
 
     do {
       chunkBytes = this._pusher(pool, bytesRead, pool.length - bytesRead);
@@ -504,33 +527,49 @@ exports.createSecurePair = function(credentials,
  * Because it is also called everywhere, we also check if the connection has
  * completed negotiation and emit 'secure' from here if it has.
  */
-SecurePair.prototype._cycle = function() {
+SecurePair.prototype._cycle = function(depth) {
+  depth = depth ? depth : 0;
   if (this._done) {
     return;
   }
 
-  // Make this function reentrant.
-  if (this._cycleLock) return;
-  this._cycleLock = true;
-  this._writeCalled = false;
+  if(depth == 0) this._writeCalled = false;
 
   var established = this._secureEstablished;
 
-  this.encrypted._pull();
-  this.cleartext._pull();
-  this.cleartext._push();
-  this.encrypted._push();
+  if (!this._cycleEncryptedPullLock) {
+    this._cycleEncryptedPullLock = true;
+    this.encrypted._pull();
+    this._cycleEncryptedPullLock = false;
+  }
 
-  this._cycleLock = false;
+  if (!this._cycleCleartextPullLock) {
+    this._cycleCleartextPullLock = true;
+    this.cleartext._pull();
+    this._cycleCleartextPullLock = false;
+  }
+
+  if (!this._cycleCleartextPushLock) {
+    this._cycleCleartextPushLock = true;
+    this.cleartext._push();
+    this._cycleCleartextPushLock = false;
+  }
+
+  if (!this._cycleEncryptedPushLock) {
+    this._cycleEncryptedPushLock = true;
+    this.encrypted._push();
+    this._cycleEncryptedPushLock = false;
+  }
 
   if (this._done) {
     return;
   }
 
-  if ((!established && this._secureEstablished) || this._writeCalled) {
+  if ((!established && this._secureEstablished) ||
+      (depth == 0 && this._writeCalled)) {
     // If we were not established but now we are, let's cycle again.
     // Or if there is some data to write...
-    this._cycle();
+    this._cycle(depth + 1);
   }
 };
 
@@ -770,6 +809,8 @@ Server.prototype.setOptions = function(options) {
 //    s.end("hello world\n");
 //  });
 //
+//
+// TODO:  make port, host part of options!
 exports.connect = function(port /* host, options, cb */) {
   // parse args
   var host, options = {}, cb;
