@@ -1,3 +1,24 @@
+// Copyright Joyent, Inc. and other Node contributors.
+//
+// Permission is hereby granted, free of charge, to any person obtaining a
+// copy of this software and associated documentation files (the
+// "Software"), to deal in the Software without restriction, including
+// without limitation the rights to use, copy, modify, merge, publish,
+// distribute, sublicense, and/or sell copies of the Software, and to permit
+// persons to whom the Software is furnished to do so, subject to the
+// following conditions:
+//
+// The above copyright notice and this permission notice shall be included
+// in all copies or substantial portions of the Software.
+//
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
+// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
+// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
+// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
+// USE OR OTHER DEALINGS IN THE SOFTWARE.
+
 exports.parse = urlParse;
 exports.resolve = urlResolve;
 exports.resolveObject = urlResolveObject;
@@ -7,11 +28,32 @@ exports.format = urlFormat;
 // compiled once on the first module load.
 var protocolPattern = /^([a-z0-9]+:)/,
     portPattern = /:[0-9]+$/,
-    nonHostChars = ['/', '?', ';', '#'],
+    delims = ['<', '>', '"', '\'', '`', /\s/],
+    unwise = ['{', '}', '|', '\\', '^', '~', '[', ']', '`'].concat(delims),
+    nonHostChars = ['/', '?', ';', '#'].concat(unwise),
+    hostnameMaxLen = 255,
+    hostnamePartPattern = /^[a-z0-9][a-z0-9A-Z-]{0,62}$/,
+    unsafeProtocol = {
+      'javascript': true,
+      'javascript:': true
+    },
     hostlessProtocol = {
+      'javascript': true,
+      'javascript:': true,
       'file': true,
       'file:': true
     },
+    pathedProtocol = {
+      'http': true,
+      'https': true,
+      'ftp': true,
+      'gopher': true,
+      'file': true,
+      'http:': true,
+      'ftp:': true,
+      'gopher:': true,
+      'file:': true
+    },
     slashedProtocol = {
       'http': true,
       'https': true,
@@ -29,7 +71,7 @@ var protocolPattern = /^([a-z0-9]+:)/,
 function urlParse(url, parseQueryString, slashesDenoteHost) {
   if (url && typeof(url) === 'object' && url.href) return url;
 
-  var out = { href: url },
+  var out = {},
       rest = url;
 
   var proto = protocolPattern.exec(rest);
@@ -50,6 +92,7 @@ function urlParse(url, parseQueryString, slashesDenoteHost) {
       out.slashes = true;
     }
   }
+
   if (!hostlessProtocol[proto] &&
       (slashes || (proto && !slashedProtocol[proto]))) {
     // there's a hostname.
@@ -79,9 +122,36 @@ function urlParse(url, parseQueryString, slashesDenoteHost) {
     // we've indicated that there is a hostname,
     // so even if it's empty, it has to be present.
     out.hostname = out.hostname || '';
+
+    // validate a little.
+    if (out.hostname.length > hostnameMaxLen) {
+      out.hostname = '';
+    } else {
+      var hostparts = out.hostname.split(/\./);
+      for (var i = 0, l = hostparts.length; i < l; i++) {
+        var part = hostparts[i];
+        if (!part.match(hostnamePartPattern)) {
+          out.hostname = '';
+          break;
+        }
+      }
+    }
   }
 
   // now rest is set to the post-host stuff.
+  // chop off any delim chars.
+  if (!unsafeProtocol[proto]) {
+    var chop = rest.length;
+    for (var i = 0, l = delims.length; i < l; i++) {
+      var c = rest.indexOf(delims[i]);
+      if (c !== -1) {
+        chop = Math.min(c, chop);
+      }
+    }
+    rest = rest.substr(0, chop);
+  }
+
+
   // chop off from the tail first.
   var hash = rest.indexOf('#');
   if (hash !== -1) {
@@ -99,9 +169,17 @@ function urlParse(url, parseQueryString, slashesDenoteHost) {
     rest = rest.slice(0, qm);
   } else if (parseQueryString) {
     // no query string, but parseQueryString still requested
+    out.search = '';
     out.query = {};
   }
   if (rest) out.pathname = rest;
+  if (slashedProtocol[proto] &&
+      out.hostname && !out.pathname) {
+    out.pathname = '/';
+  }
+
+  // finally, reconstruct the href based on what has been validated.
+  out.href = urlFormat(out);
 
   return out;
 }
@@ -123,13 +201,12 @@ function urlFormat(obj) {
           ) :
           false,
       pathname = obj.pathname || '',
-      search = obj.search || (
-          obj.query && ('?' + (
-              typeof(obj.query) === 'object' ?
-              querystring.stringify(obj.query) :
-              String(obj.query)
-          ))
-      ) || '',
+      query = obj.query &&
+              ((typeof obj.query === 'object' &&
+                Object.keys(obj.query).length) ?
+                 querystring.stringify(obj.query) :
+                 '') || '',
+      search = obj.search || (query && ('?' + query)) || '',
       hash = obj.hash || '';
 
   if (protocol && protocol.substr(-1) !== ':') protocol += ':';
